#    Copyright (C) 2017 SUSE LLC
#
#    Author: Robert Schweikert <rjschwei@suse.com>
#
# This file is part of cloud-init. See LICENSE file for license information.

from cloudinit.distros import opensuse


class Distro(opensuse.Distro):
    pass
